{% extends "!layout.html" %}

{% block footer %}
{{ super() }}
<!-- Cloudflare Web Analytics --><script defer src='https://static.cloudflareinsights.com/beacon.min.js' data-cf-beacon='{"token": "b5047b826dff4e4db4b9193bb47eec37"}'></script><!-- End Cloudflare Web Analytics -->
{% endblock %}