.highlight.lmql {
    border: 1pt solid rgb(212, 212, 212);
    position: relative;
}

html[data-theme="dark"] .highlight.lmql {
    border: 1pt solid rgb(110, 109, 109);
}

.highlight.lmql iframe {
    background-color: var(--pst-color-background);
}

.highlight.lmql.playground {
    border: none;
}

.highlight.lmql pre {
    border: none;
}

.highlight pre {
    padding-top: 20pt !important;
    padding-bottom: 20pt !important;
}

.highlight.lmql button {
    background: transparent;
    border: none;

    text-align: right;
    font-size: 0.8em;
    font-weight: regular;
    padding: 0;
    position: absolute;
    top: 5pt;
    right: 5pt;
    color: var(--pst-color-link);
}

.highlight.lmql button:hover {
    cursor: pointer;
    text-decoration: underline;
}

.getting-started-options {
    margin: 0 0;
    max-width: 800px;
}

.getting-started:hover {
    border: 0.5pt solid rgb(143, 143, 143);
}

.getting-started {
    width: 170pt;
    max-width: calc(100% - 20pt);
    font-size: 0.85em;
    border: 0.5 pt solid grey;
    padding: 5pt;
    margin: 5pt 4pt;
    border: 0.5pt solid rgb(193, 190, 190);
    
    border-radius: 2pt;
    display: flex;
    flex-direction: row;
    flex-direction: column;
    box-shadow: 0pt 0pt 10pt 0pt rgba(0, 0, 0, 0.1);
}

@media (max-width: 540pt) {
    .getting-started {
        width: 100%;
    }
}

@media screen and (max-width: 769px), print {
    .getting-started {
        margin-left: auto;
        margin-right: auto;
    }
}

.getting-started h2 {
    margin: 5pt 0pt;
}

.getting-started>a {
    display: block;
    padding: 8pt;
    border-radius: 4pt;
    background-color: rgb(255, 255, 255);
    margin: 2pt 2pt;
    border: 1pt solid rgb(223, 218, 218);
    font-size: 10pt;
    font-weight: bold;
    font-family: system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, Cantarell, 'Open Sans', 'Helvetica Neue', sans-serif;
    cursor: pointer;
    color:rgba(0, 0, 0, 0.957)
}

html[data-theme="dark"] .getting-started>a {
    background-color: rgb(57, 57, 57) !important;
    border-color: rgb(75, 75, 75);
    color: rgb(215, 215, 215);
}

html[data-theme="dark"] .getting-started>a.primary {
    background-color: rgb(66, 64, 174) !important;
}

.getting-started>a:hover {
    background-color: rgb(233, 228, 228);
}


.getting-started .cmd {
    padding: 5pt;
    margin: 5pt auto;
    max-width: 280pt;
    border-radius: 4pt;
    background-color: #dcdcdc;
    color: rgb(35, 35, 35);
    font-family: monospace;
    font-weight: bold;

    /* on double click select all */
    -webkit-user-select: all;
    -moz-user-select: all;
    -ms-user-select: all;
    user-select: all;
    cursor: pointer;
}

.getting-started>span>a {
    display: inline;
}

.getting-started>a.primary {
    background-color: rgb(107, 119, 255);
    color: white;
}

.getting-started>a.primary:hover {
    background-color: rgb(122, 132, 245);
    color: white;
}

.columns.is-getting-started {
    margin-top: 5pt;    
}

.columns {
    display: flex;
    flex-wrap: wrap;
    flex-direction: row;
    text-align: center;
}

.columns.is-getting-started h2 {
    font-size: 10pt;
    text-align: center;
}

.nboutput .highlight pre {
    /* break lines */
    white-space: break-spaces !important;
}

.highlight-model-output::before {
    content: "Model Output";
    font-family: system-ui, -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Oxygen, Ubuntu, Cantarell, 'Open Sans', 'Helvetica Neue', sans-serif;
    position: absolute;
    top: 4pt;
    font-size: 9pt;
    text-transform: uppercase;
    left: 0pt;
    z-index: 2;
    width: 100%;
    text-align: center;
    color: grey;
}

.highlight-model-output {
    position: relative;
    padding-left: 5pt;
    top: -5pt;
    padding-right: 5pt;
    z-index: -1;
    font-size: 0.9em;
    font-family: monospace;
}

html[data-theme="dark"] .highlight-model-output::before {
    color: rgb(143, 143, 143);
}

html[data-theme="dark"] .highlight-model-output {
    background: transparent;
}

.highlight-model-output .highlight, html[data-theme="dark"] .highlight-model-output .highlight {
    display: block;
    white-space: break-spaces !important;
}


.highlight-model-output .highlight {
    background-color: #fbfbfb !important;
    border: 0.5pt solid grey;
    position: relative;
    /* top: -12pt; */
    border-color: rgb(223, 219, 219);
    white-space: break-spaces !important;
    padding: 10pt;
    padding-top: 20pt;
}

html[data-theme="dark"] .highlight-model-output .highlight {
    background-color: rgb(30, 30, 30) !important;
    border-color: rgb(68, 68, 68) !important;
}

.variable {
    display: inline; 
    padding: 0.5pt;
}

.variable .variable-name {
    font-weight: 700;
    color: #333;
    background-color: rgba(0, 0, 0, 0.4);
    border-radius: 0.2em;
    font-family: monospace;
    color: rgba(255, 255, 255, 0.9);
    display: inline;
    margin: 0;
    margin-right: 0.0em;
    margin-left: 0;
    font-size: 0.7em;
    padding: 0.12em;
    position: relative;
    top: -0.15em;
    left: 0.1em;

    transform-origin: 0 50%;
    overflow: hidden;
}

.variable .variable-name.visible {
    opacity: 1.0;
}

.val1 {
    background-color: rgba(255, 166, 0, 0.5) !important;
    cursor: default;
}

.val1:hover, .val1.hover, .val1-hover .val1 {
    background-color: rgba(255, 166, 0, 1.0) !important;
}
.val2 {
    background-color: rgba(206, 128, 184, 1.0) !important;
    cursor: default;
}

.val2:hover, .val2.hover, .val2-hover .val2 {
    background-color: rgba(206, 128, 184, 1.0) !important;
}
.val3 {
    background-color: rgb(130, 140, 249) !important;
    cursor: default;
}

.val3:hover, .val3.hover, .val3-hover .val3 {
    background-color: rgba(107, 119, 255, 1.0) !important;
}