def gsm8k_samples(): return """Q: Every hour Joanne has to collect the coins out of the fountain inside the mall. During the first hour, she collected 15 coins. For the next two hours, she collected 35 coins from the fountain. In the fourth hour, she collected 50 coins from the fountain but she gave 15 of them to her coworker so she could buy a soda. How many coins did she have after the fourth hour?
A: Let's think step by step.
15 coins collected in hour one
35 coins collected in hour two
35 coins collected in hour three
50 coins collected in hour four
Before giving her coworker some coins there were 15+35+35+50=<< 15+35+35+50 = 135 >>135 coins
The number of coins after given 15 to her coworker is 135-15=<< 135-15 = 120 >>120
So the answer is 120

Q: Jerry’s two daughters play softball on different teams. They each have 8 games this season. Each team practices 4 hours for every game they play. If each game lasts for 2 hours, how many hours will Jerry spend at the field watching his daughters play and practice altogether?
A: Let's think step by step.
Jerry will spend 8 games x 2 hours per game = << 8*2 = 16 >>16 hours watching one daughter play her games.
He will spend 16 x 2 = << 16*2 = 32 >>32 hours watching both daughters play their games.
He will spend 8 games x 4 hours of practice = << 8*4 = 32 >>32 hours watching one daughter practice.
He will spend 32 x 2 = << 32*2 = 64 >>64 hours watching both daughters practice.
He will spend a total of 32 hours watching games + 64 hours watching practice = << 32+64 = 96 >>96 hours.
So the answer is 96
"""
