import styled from 'styled-components';
import Editor from "@monaco-editor/react";
import React, { useEffect, useRef, useState } from "react";
import { registerLmqlLanguage } from "./editor/lmql-monaco-language";
import { BsTerminal, BsFileCode } from 'react-icons/bs';

export function DecodingTree(props) {
    return <div/>;
}